#!/usr/bin/perl

# exiscan 0.9 by tom kistner <tom@duncanthrax.net>
#
# Thanks to
# 	- Marcus Klein <marcus@netline-ag.de> (AVP support)
#	- Chris Beauchamp <chris@buildstore.co.uk> (Sophos Support)
#	- Rick Ballard <Richard.Ballard@xwave.com> (some good suggestions ;)
#
# URL: http://duncanthrax.net/exiscan/

# !!! see INSTALL for installation instructions first !!!

# **********************************************************************
# CONFIGURE exiscan HERE
# This is the configuration section. Take the time to read the comments.
# You will probably need a second terminal to find things out, so open 
# one now :)

	# ------------------------------------------------------------------
	# Path and program locations
	#
	# exiscan uses several external binaries to do its job. Most of them
	# are standard unix utilities which should be present on every decent
	# installation. A download URL is given for the more exotic ones.

	# this is where you have put the exiscan package

	$basepath="/usr/exim/exiscan";


	# exims queue directory

	$queuedir="/usr/exim/spool/input";


	# the "mail" executable. used to send simple notifications
	# to senders and recipients (these features can be switched 
	# on/off further below)

	$mail="/bin/mail";


	# the reformime executable. the preferred MIME mail unpacker.
	# this is part of the "maildrop" package (check FreshMeat on
	# "maildrop"). 
	# See http://duncanthrax.net/exiscan/ for a 
	# precompiled reformime binary for linux glibc systems.

	$reformime = "/usr/local/bin/reformime";


	# the metamail executable. used to unpack mails.
	# almost present on any unix system, but very
	# old and not very reliable. if you have defined
	# the "reformime" entry above then you will NOT
	# need to change this.

	$metamail="<NOT USED>";


	# the tnef executable (OPTIONAL)
	# get it at http://world.std.com/~damned/software.html
	# a very neat utility which can decode Micro$oft's
	# MS-TNEF "wrapper" format, typically created by the
	# evil EXCHANGE or OUTLOOK programs. YUKK !!
	# This will enable exiscan to look into these obsfucated
	# mails :) Set this to "0" if you do not want tnef support.

	$tnef = "/usr/local/bin/tnef";


	# metasend, to forward the virus infected mails as an attach to
	# the admin.
	# WARNING: one user reported that his copy of metasend which came
	# with Slackware 7 was broken. Simply replacing it with a version
	# from a recent redhat system fixed it.

	$metasend="/usr/bin/metasend";


	# the virus scanner to use ("mcafee", "trendmicro",
	# "sophos" or "avp")

	$scanner="trendmicro";


	# the virus scanner executable (full path + executable)

	$scannerex="/etc/iscan/vscan";


	# the exim executable

	$exim="/usr/exim/bin/exim";



	# ------------------------------------------------------------------
	#
	# Security and Features configuration

	# the crypt() salt
	# set this to a 2-character string of your liking.
	# it is used to determine if a message was REALLY already scanned 
	# by our system. This prevents attackers from faking a X-Scanner line 
	# in the mail header. Yes, i KNOW that this is not 100% security, but
	# unless you are the admin of Fort Knox it should be sufficient.
	
	$salt = "fo";


	# exiscan "forgets" about already scanned mail ID's every $resetinterval 
	# seconds. This should be set to the value of the exim dequeue interval
	# (if you ran exim with a flag of -q1h, then you should set "3600" here.)

	$resetinterval="3600";


	# double-unpack SMIME wrapped messages ? such messages are usually
	# "double-wrapped" either in MIME or MS-TNEF so we need a second
	# unpack run. If you want max speed, set this to 0.
	# If you want more security, set this to 1 (default).

	$unpack_smime = 1;


	# the email addy that gets notification about found viruses. mind the escaped '@' !

	$postmaster="postmaster\@eurescom.de";


	# set this to "1" if you want to notify the sender of a virus infected mail
	# "0" switches the feature off

	$sender_notification = 1;


	# set this to "1" if you want to notify the recipients of a virus infected mail
	# "0" switches the feature off

	$rcpt_notification = 1;


	# Customize the text used in the notification to the sender and recipients.
	# You can use the following variables:
	# \n -> a newline
	# \$from -> the sender of the message (<sender@domain.com>)
	# \@to -> space separated list of recipients
	# \$subject -> the subject. may contain funky quoted-printable characters.
	# \$postmaster -> the postmaster's email address
	# The "\" is necessary as this text will be eval'ed later ..
	# the "return" must also stay in place ...
	#
	# WARNING: if you don't know what you are doing here,
	# please leave the default texts in place. Thanks :)

	$sender_notification_text = "return \"Your EMail with subject '\$subject', sent to the recipient(s)\n\n\@to\n\ncontains a virus or other harmful content. The message has NOT been delivered to the recipients.\nPlease contact the postmaster (mailto:\$postmaster) to resolve this issue.\"";
	$rcpt_notification_text = "return \"An EMail directed to you, with subject '\$subject',\nfrom \$from, contained a virus or other harmful content.\nDo NOT reply to this e-mail.\nContact the original sender of the message (mailto:\$from) to resolve this issue.\"";


	# the syslog facility to use. Leave it as it is to get output in /var/log/messages.
	$facility = "LOG_LOCAL0";


	# ------------------------------------------------------------------
	#
	# Optional Path Configuration
	# You may adjust these settings too but it's not necessary if you use
	# the standard exiscan directory layout

	# directory where mails are unpacked
	
	$checkdir="$basepath/checkqueue";


	# where to put virus infested mails
	
	$virusdir="$basepath/virusmails";


	# ------------------------------------------------------------------
	#
	# Debug mode. When this is set to 1, exiscan will not fork to the
	# background. It will rather stay on the terminal and output lots of
	# informational data. If you have any problems with exiscan then
	# switch this on and maybe send me the last output lines ...

	$debug = 0;


# end of configuration section
# **********************************************************************
# no editing below this line ... unless you know what you are doing ;)


use File::Copy;
use File::Remove qw(remove);

# chdir to base path, so core files will go into the right directory <g>
chdir("$basepath");

# now some sanity checks

print "Checking for mail .. ";
if (!(-x $mail)) {
	print "not found ($mail) ! Please check your settings\n";
	exit(1);
};
print "found\n";


print "Checking for reformime .. ";
if (!(-x $reformime)) {
	print "not found. Falling back to metamail.\n";
	$reformime = 0;
	print "Checking for metamail .. ";
	if (!(-x $metamail)) {
		print "not found ($metamail) ! Please check your settings\n";
		exit(1);
	};
	print "found\n";
}
else {
	print "found, using reformime to unpack MIME mails.\n";
};

print "Checking for metasend .. ";
if (!(-x $metasend)) {
        print "not found ($metasend) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for tnef .. ";
if (!(-x $tnef)) {
        print "not found, will not unpack MS-TNEF format\n";
        $tnef = 0;
} 
else {
	print "found\n";
};

print "Checking for virus scanner .. ";
if (!(-x $scannerex)) {
        print "not found ($scannerex) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for exim .. ";
if (!(-x $exim)) {
        print "not found ($exim) ! Please check your settings\n";
        exit(1);
};
print "found\n";

$quit = 0;
$| = 0;

# fork a child and exit, if debug mode is not enabled

if (!($debug)) {
	FORK: {
		if ($pid = fork) {
			print "exiscan launched into the background (pid $pid)\n";
			exit(0);
		}
		elsif (defined $pid) {
			goto CHILD;
		}
		else {
			die "exiscan can't fork to the background: $!\n";
		};
	};
} 
else {
	print "DEBUG mode enabled, running on terminal, CTRL-C to abort\n";
};

CHILD:

%checked = ();
$lastreset = time();

# ---------------------------------
# the exit handler
# HUP signal makes exiscan quit
# gracefully

sub catch_zap {
       my $signame = shift;
       $quit = 1;        
}

$SIG{HUP} = \&catch_zap;

# --------------------------------------
# set up syslog

use Unix::Syslog qw(:macros);
use Unix::Syslog qw(:subs);

openlog "exiscan", LOG_PID, $facility;

syslog LOG_INFO, "started";


# --------------------------------------
# the superloop
# it is in an eval block to trap unexpected errors.
# the handler will log an error with the mail ID
# and restart the superloop
# setting $quit to 1 (via HUP signal) is the only 
# way to exit this superloop correctly ...

SUPERLOOP:
eval { 
print "Entered superloop.\n" if ($debug);
while(!($quit)) {

# --------------------------------------
# forget scanned mails IDs every interval
# to allow re-dequeueing

 if ((time() - $lastreset) > $resetinterval) {
	 print "ID cache cleared.\n" if ($debug);
 	 syslog LOG_INFO, " exiscan clearing ID cache\n";
	 $lastreset = time();
	 %checked = ();
 };

# --------------------------------------
# get the contents of the queue (headers)

 opendir(QUEUEDIR,"$queuedir");
 @mails = grep /\-H/, readdir QUEUEDIR;
 closedir(QUEUEDIR);
 print "Scanned $queuedir and found this: @mails\n" if ($debug);

# --------------------------------------
# once second grace to decrease load
# and allow exim to perhaps finish
# writing a header file

 sleep 1;

# ---------------------------------------
# process all found mails

 foreach $mheader (@mails) {

	if ($quit) { goto EXITEXISCAN };

	$id = $mheader; $id =~ s/\-H$//g;

	print "Entering process loop for $id\n" if ($debug);

	if ( (-e "$queuedir/$id-H") && (!($checked{$id})) ) {

               # -------------------------------------
               # check for X-Scanner in the header

		open(HEADER,"< $queuedir/$id-H");
		if ((@scanner = grep /082  X-Scanner\: exiscan/, <HEADER>)) {
			
			close(HEADER);

			print "$id: header read\n" if ($debug);

		       # ------------------------------------
                       # see if WE did really scan it
		       # we don't trust anyone else ...
			
			($rest,$oldid,$cryptid,$rest2) = split /\*/, @scanner[$#scanner], 4;
			$checkid = crypt($oldid,$salt);
			$checkid =~ s/^..//;

			if ($checkid eq $cryptid) {

				if ($checked{$oldid}) {
					print "$id: resent, prev ID $oldid\n" if ($debug);
					syslog LOG_INFO, "ID $id R:resent, prev ID $oldid, dequeuing";
				}
				else {
					print "$id: delayed, attempting dequeue again\n" if ($debug);				
					syslog LOG_INFO, "ID $id R:delayed, attempting dequeue again";
				};
			
				$checked{$id} = 1;
			
                       	       # ------------------------------------
                               # de-queue 
                      
				print "$id: launching exim dequeue process\n" if ($debug);
				system("$exim -Mc $id &");


                               # ------------------------------------
                               # process next mail

				print "$id: leaving process loop\n" if ($debug);
				next;
			
			}
			else {
				print "$id: last X-Scanner line not from us\n" if ($debug);
				syslog LOG_INFO, "ID $id R: last X-Scanner line is not from us. Scanning now.";
			};
		}
		else {
			print "$id: this is a new one\n" if ($debug);
			close(HEADER);
		};

	       # -------------------------------------
	       # build complete mail in standard format
	       # get real TO and FROM addresses
	       
		# void the variables, just for paranoia
		@completemail = ();
		$from = "<EMPTY>";
		@to = ();
		$i = "";
		$subject = "<EMPTY>";
		
		open(HEADER,"< $queuedir/$id-H");
		do {
			if ($i =~ /^\</) { $from = $i; chop($from); };		
			if ( ($i =~ /.+\@.+/) && (!($i =~ /^\</)) && (!($i =~ /^.. /)) && (!($i =~ /^\-/)) ) {
				$tmp = $i; chop($tmp);
				push @to, $tmp;
			};
			$i = <HEADER>;
		} while (!($i eq "\n"));
		
		while(<HEADER>) {
			# grab the subject if we find one
			if ($_ =~ /Subject\: /) {
				$tmp = $_; chop($tmp);
				($dummy,$subject) = split /Subject\: /, $tmp, 2;
			};
			
			# kick out old exiscan X-Scanner lines
			next if ($_ =~ /082  X-Scanner\: exiscan/);

			$_ =~ s/^.....//g if ($_ =~ /^[0-9][0-9][0-9]/);
			push @completemail, $_;
		};
		close(HEADER);

	       # ------------------------------------
	       # make sure the @to array does not get too large
	       # this works around a buffer overflow which obviously 
	       # appears when Unix::Syslog is logging to older syslogds
	       # right now we set the limit to 5 addresses plus a
	       # "truncated" message

		if ($#to > 5) {
			$#to = 5;
			push @to, "<more recipients, but not shown>";
		};	

	       # -------------------------------------
	       # do not scan the mail if it is sent to
	       # the postmaster address ONLY

		if ( ($#to == 0) && (@to[0] eq $postmaster) ) {
			print "$id: postmaster-only mail\n" if ($debug);
			syslog LOG_INFO, "ID $id R:PostMaster mail only, dequeuing.";
			
			# dequeue
			$checked{$id} = 1;
                        system("$exim -Mc $id &");
		
			# process next mail
			print "$id: leaving process loop\n" if ($debug);
			next;
		};

	       # -------------------------------------
	       # add our tag to the header
		
		$crypttag = crypt($id,$salt);
		# cut away the salt.
		$crypttag =~ s/^..//;
		push @completemail, "X-Scanner: exiscan \*$id\*$crypttag\* http://duncanthrax.net/exiscan/\n";

		# add separator
		push @completemail, "\n";

	       # -------------------------------------
	       # append the complete body

		open(BODY,"< $queuedir/$id-D");
                $i = <BODY>;	# kick out 1st line
		push @completemail, <BODY>;
                close(BODY);

	      # -------------------------------------
	      # write the complete mail to disk.
	      # that gives us a file in standard MBOX format

		open(TMPMAIL,"> $checkdir/$id-complete");
		print TMPMAIL @completemail;
		close(TMPMAIL);

               # -------------------------------------
               # unpack mail in temporary directory

		mkdir "$checkdir/$id-tmp", 0755;
		
		print "$id: unpacking in $checkdir/$id-tmp\n" if ($debug);
		# use reformime if present
		if ($reformime) {
			chdir("$checkdir/$id-tmp");
			system("$reformime -xf- < $checkdir/$id-complete");
			chdir("$basepath");
		}
		# else we use good 'ol metamail
		else {
			system("METAMAIL_TMPDIR=$checkdir/$id-tmp; export METAMAIL_TMPDIR; $metamail -r -q -x -w $checkdir/$id-complete > /dev/null 2> /dev/null");
		};

	       # -------------------------------------
	       # enumerate the unpacked files and feed
	       # winmail.dat (MS-TNEF) to tnef, if configured

		if ($tnef) {
			opendir(CHECKDIR,"$checkdir/$id-tmp");
			@mimefiles = grep /winmail/i, readdir CHECKDIR;
			closedir(CHECKDIR);

			print "$id: scanned $checkdir/$id-tmp and found TNEF: @mimefiles\n" if ($debug);		

			foreach $mimefile (@mimefiles) {
				system("$tnef -f $checkdir/$id-tmp/$mimefile -C $checkdir/$id-tmp --overwrite");
			};
		};

	       # -------------------------------------
	       # attempt to unpack SMIME encapsulated messages
	       # these will typically come out of MS-TNEF encapsulated messages

		if ($unpack_smime) {
			opendir(CHECKDIR,"$checkdir/$id-tmp");
			@mimefiles = grep /smime/i, readdir CHECKDIR;
			closedir(CHECKDIR);

			print "$id: scanned $checkdir/$id-tmp and found SMIME: @mimefiles\n" if ($debug);
                
			foreach $mimefile (@mimefiles) {
				# use reformime if present
				if ($reformime) {
	                       		chdir("$checkdir/$id-tmp");
			       		system("$reformime -xf- < $checkdir/$id-tmp/$mimefile");
		  			chdir("$basepath");  
                		}
		 		# else we use good 'ol metamail
                		else {
                        		system("METAMAIL_TMPDIR=$checkdir/$id-tmp; export METAMAIL_TMPDIR; $metamail -r -q -x -w $checkdir/$id-tmp/$mimefile > /dev/null 2> /dev/null");
				};
	        	};
		};

               # -------------------------------------
               # run virus scanner and get result

		$result = 0;

		# McAfee's uvscan
		if ($scanner =~ /mcafee/i) {
			$result = system("$scannerex -r $checkdir/$id-tmp --unzip > /dev/null 2> /dev/null");
		};

		# Trend Micro's vscan
		if ($scanner =~ /trendmicro/i) {
			open(VSCAN,"$scannerex $checkdir/$id-tmp |");

			while(<VSCAN>) {
			        if ($_ =~ /Infected \: /) {
		        	        chop;
		                	($dummy,$infected) = split / \: /,$_,2;
			                if ($infected =~ /[1-9]/) {
		        	                $result = 1;
        		        	        last;
		        	        };
			        };
			};
			close(VSCAN);
		};

		# Sophos Sweep
		if ($scanner =~ /sophos/i) {
			$result = system("$scannerex sweep -archive -f -all -rec $checkdir/$id-tmp > /dev/null 2> /dev/null");
		};

		# Kapersky's AVP
		if ($scanner =~ /avp/i) {
			open(VSCAN,"$scannerex $checkdir/$id-complete |");
			
			while(<VSCAN>) {
				if ($_ =~ /Return code/) {
					chop;
					($dummy,$infected) = split /\: /,$_,2;
					if ($infected =~ /[1-8]/) {
						$result = 1;
						last;
					};
				};
			};
			close(VSCAN);
		};

		if ($result) {
                
		       # ------------------------------------
                       # Virus found
			
			print "$id: virus found\n" if ($debug);

			syslog LOG_INFO, "ID $id F:$from T:@to R:virus found, moving to $virusdir";
	
			move("$queuedir/$id-H","$virusdir/$id-H");
			move("$queuedir/$id-D","$virusdir/$id-D");


		       # ------------------------------------
		       # send notification to admin. Infected mail is attached.
	
			syslog LOG_INFO, "ID $id sending mail attachment to $postmaster";
			system("$metasend -t $postmaster -F exiscan -s \"Virus in mail $id\" -m message/rfc822 -f $checkdir/$id-complete -S 99999999 -b -e 7bit");


			# -----------------------------------
			# send notification to sender (if configured)

			if ($sender_notification) {
				syslog LOG_INFO, "ID $id sending notification to sender ($from)";
				my $cleanfrom = $from; $cleanfrom =~ s/\<|\>//g;
				open(SENDER,"| $mail -s \"Virus\ /\ harmful\ content\ found\ in\ EMail\" $cleanfrom"); 
				print SENDER eval $sender_notification_text;
				close(SENDER);
			};


			# -----------------------------------
			# send notification to recipients (if configured)

			if ($rcpt_notification) {
				syslog LOG_INFO, "ID $id sending notification to recipients (@to)";
				open(RCPT,"| $mail -s \"Virus\ /\ harmful\ content\ found\ in\ EMail\" @to"); 
				print RCPT eval $rcpt_notification_text;
				close(RCPT);
			};
		}
		else {
		
		       # ------------------------------------
		       # insert X-Scanner header line	

			$crypttag = crypt($id,$salt);
			# cut away the salt.
			$crypttag =~ s/^..//;

			if (-e "$queuedir/$id-H") {
				open(HEADER,">> $queuedir/$id-H");
				print HEADER "082  X-Scanner: exiscan \*$id\*$crypttag\* http://duncanthrax.net/exiscan/\n";
				close(HEADER);
			};

                       # ------------------------------------
                       # remember mail id

			$checked{$id} = 1;

                       # ------------------------------------
                       # Mail is ok, de-queue

			print "$id: clean, launching exim dequeue process\n" if ($debug);
			syslog LOG_INFO, "ID $id F:$from T:@to R:mail clean, dequeuing"; 
			system("$exim -Mc $id &");
		};

               # -------------------------------------
               # clean up temp files

		print "$id: clean up temp files/dirs\n" if ($debug);
		remove \1, "$checkdir/$id-tmp";
		unlink "$checkdir/$id-complete";
		
		print "$id: leaving process loop\n" if ($debug);
	};
	print "$id: already worked on it since last hash reset, processing next\n" if ($debug);
   };
 };
 };

# -----------------------------------
# if $quit is not 1 then this code got 
# reached when the superloop eval block died 
# unexpectedly. we try to output some meaningful
# error message to the log and simply restart
# the superloop ..

if (!($quit)) {
	print "$id: error ($@). Last mail ID was $id.\n" if ($debug);
	syslog LOG_INFO, "error occured ($@). Last mail ID was $id.";
	goto SUPERLOOP;
};

# -----------------------------------
# the exit routine. gets passed
# when we shut down with SIGHUP recvd

EXITEXISCAN:

print "$id: terminating with HUP signal\n" if ($debug);
syslog LOG_INFO, "terminating";
closelog;
exit(0);
